using System;

namespace Printy.Library
{
  /// <summary>
  /// Konfigurationsklasse fr einen Printer.
  /// In dieser Klasse wird auch verwaltet, wie die Abstnde zwischen den Symbolen aussehen.
  /// Die Standardeinstellung ist die folgende:
  /// - zwischen aufeinander folgenden Symbolen und Klammerungen gibt es ein Zeichen Abstand;
  /// - vor und nach '=' sowie '|' gibt es ein Zeichen Abstand;
  /// - nach ffnender und vor schlieender Klammer; Ausnahme: leere Klammerung
  /// - Kommentare werden ebenfalls mit einem Zeichen Abstand ausgegeben;
  /// - sonst werden keine Abstnde eingefgt;
  /// - ein Zeilenwechsel auf Grund der Spaltenbreite ist immer erlaubt;
  /// Dieses Verhalten kann umfassend konfiguriert werden.
  /// Auch das Verhalten bei Einzgen kann mit den Bitvektoren openSet und closeSet
  /// angepasst werden.
  /// </summary>
  public class Config
  {
    /// <summary>
    /// Klassenname fr jeden Typ
    /// </summary>
    public static readonly string[] Names = 
    {
      "",
      "Production",
      "Symbol",
      "Literal",
      "Open",
      "Close",
      "Assign",
      "Separator",
      "Delimiter",
      "InnerComment",
      "OuterComment",
      "Attribute",
      "Resolver",
      "Prolog",
      "Epilog",
      "Action",
      "LineNumber"
    };

    /// <summary>
    /// Zusammengefasste Symbol-Klassen.
    /// Diese Konstanten werden fr PutSpace() bentigt.
    /// </summary>
    public const int
      ALL          =-1,  // Konstante, die alle Symbole als Menge reprsentiert.
      INDENT       = 0,  // Einzug
      Production   = 1,  // Produktion
      Symbol       = 2,  // Symbol (TS oder NTS)
      Literal      = 3,  // Literal (string | char | "WEAK" | "SYNC" | "ANY")
      Open         = 4,  // '(' | '[' | '{' sowie ffnende Klammerung von Alternativenlisten
      Close        = 5,  // ')' | ']' | '}' sowie schlieende Klammerung von Alternativenlisten
      Assign       = 6,  // '='
      Separator    = 7,  // '|'
      Delimiter    = 8,  // '.'
      InnerComment = 9,  // beliebiger Kommentar innerhalb einer Produktion
      OuterComment = 10, // beliebiger Kommentar zwischen Produktionen
      Attribute    = 11, // Attributliste <...>
      Resolver     = 12, // LL(1) conflict resolver IF (...)
      Prolog       = 13, // Prolog
      Epilog       = 14, // Epilog
      Action       = 15, // semantische Aktion
      LineNumber   = 16; // Zeilennummer

    /// <summary>
    /// Wert fr altPar: Alternativen niemals klammern
    /// </summary>
    public const int ALT_PAR_NEVER = STD_LEVEL;

    /// <summary>
    /// Wert fr altPar: Alternativen immer klammern
    /// </summary>
    public const int ALT_PAR_ALWAYS = MAX_LEVEL;

    /// <summary>
    /// Konstante fr PutBreak(): Zeilenumbruch zwischen zwei Symbolen immer dann,
    /// wenn es die maximale Spaltenbreite verlangt.
    /// Dieses Verhalten ist Standard fr alle Symbole.
    /// </summary>
    public const int BREAK_DEFAULT = STD_LEVEL;

    /// <summary>
    /// Konstante fr PutBreak(): Zeilenumbruch zwischen zwei Symbolen wird unterdrckt;
    /// die maximale Spaltenbreite kann dadurch berschritten werden.
    /// </summary>
    public const int BREAK_NEVER = MIN_LEVEL;

    /// <summary>
    /// Konstante fr PutBreak(): Zeilenumbruch zwischen zwei Symbolen wird erzwungen.
    /// </summary>
    public const int BREAK_ALWAYS = MAX_LEVEL;

    /// <summary>
    /// Minimaler Wert fr einen Level.
    /// </summary>
    const int MIN_LEVEL = sbyte.MinValue;

    /// <summary>
    /// Maximaler Wert fr einen Level.
    /// </summary>
    const int MAX_LEVEL = sbyte.MaxValue;

    /// <summary>
    /// Normaler Wert fr einen Level.
    /// </summary>
    const int STD_LEVEL = 0;
  
    /// <summary>
    /// Erstes Symbol in der Matrix
    /// </summary>
    const int MIN_SYMBOL = 1;

    /// <summary>
    /// Letztes Symbol in der Matrix
    /// </summary>
    const int MAX_SYMBOL = 14;

    /// <summary>
    /// Testet die Verschachtelungstiefe gegen einen Tiefenparameter.
    /// </summary>
    /// <param name="level">aktuelle Ebene</param>
    /// <param name="threshold">Test-Ebene</param>
    /// <returns>true, falls der Test erfolgreich war</returns>
    public static bool TestLevel(int level, int threshold)
    {
      if (threshold > 0)
      {
        return (threshold == MAX_LEVEL) || (level <= threshold);
      }
      if (threshold < 0)
      {
        return (threshold == MIN_LEVEL) || (level > -threshold);
      }
      return false;

    }

    /// <summary>
    /// Dieser Wert gibt die maximale Verschachtelungstiefe an, bei jener eine Aufzhlung von
    /// Alternativen geklammert wird. 
    /// Besipiel: altParLevel=2 bedeutet, dass nur auf der Ebene von Produktion Alternativen
    /// geklammert werden.
    /// Standardwert: 0 (keine Klammerungen)
    /// </summary>
    public int altPar = 0;

    /// <summary>
    /// Dieser Wert gibt jene Breite an, innerhalb welcher ffnende Symbole erlaubt sind.
    /// Jenseits dieser Breite werden sie in die nchste Zeile umgebrochen.
    /// Mit dieser Parameter kann vermieden werden, dass allzu groe Einzge entstehen,
    /// die die Spaltenbreite berschreiten.
    /// Standardwert: 33 Zeichen
    /// </summary>
    public int indentWidth = 33;

    /// <summary>
    /// Breite der linken Spalte.
    /// Dieser Wert wird niemals berschritten, es sei denn, es kommen einzelne Symbole
    /// vor, die nicht in eine Zeile passen; in diesem Fall wird die berschreitung der maximalen
    /// Breite gering gehalten, indem das extra-lange Symbol in eine eigene Zeile geschrieben wird.
    /// Standardwert: 66 Zeichen
    /// </summary>
    public int colWidth = 66;

    /// <summary>
    /// Abstand der rechten Spalte von der linken.
    /// Standardwert: 4 Zeichen
    /// </summary>
    public int colSpace = 4;

    /// <summary>
    /// Wirkung:
    /// Schlieende Symbole werden in die nchste Zeile verschoben. Somit
    /// ist gewhrleistet, dass ')' unter '(', und '.' unter '=' usw (hngt von
    /// openSet und closeSet ab) gedruckt wird.
    /// Mit diesem Parameter wird eine intelligente Einrckungsheuristik aktiviert,
    /// die die manuelle Einstellung mit PutBreak() ergnzt.
    /// Stanndardwert: false
    /// </summary>
    public bool autoBlock = false;

    /// <summary>
    /// Dieser Parameter legt fest, ob die Breite der linken Spalte automatisch bestimmt wird.
    /// Falls true, gilt columnWidth als maximale Breite der linken Spalte, falls false,
    /// gilt columnWidth als fixe Breite der linken Spalte; es sei denn, dass ein extra-langes
    /// Symbol vorkommt.
    /// Standardwert: false
    /// </summary>
    public bool autoFit = false;

    /// <summary>
    /// Diese Option deaktiviert, falls true, sogenannte Source Breaks.
    /// Dies sind Zeilenwechsel, die im Quelldokument vorkommen.
    /// Auf diese Weise kann die Gestaltung von Zeilenumbrchen vom Printy
    /// vom Quelldokument unabhngig gemacht werden.
    /// Standardwert: false
    /// </summary>
    public bool noBreaks = false;

    /// <summary>
    /// Abstandszeichen 
    /// Standardwert: ' ' (Unicode #32)
    /// </summary>
    public char spaceChar = ' ';

    /// <summary>
    /// Dies ist der Formatstring fr Zeilennummern.
    /// Falls null, werden keine Zeilennummern ausgegeben.
    /// Bsp: 
    /// config.lineNum = "(*000*) ";
    /// Dies gibt Zeilennummern innerhalb von EBNF-Kommentaren mit 3 Ziffern aus.
    /// Zeilennummern werden immer linksbndig gedruckt.
    /// Standardwert: null
    /// </summary>
    public string lineNum = null;

    /// <summary>
    /// ffnende Klammer von Alternativenlisten.
    /// Standardwert: "("
    /// </summary>
    public string altParOpen = "(";

    /// <summary>
    /// Schlieende Klammer von Alternativenlisten.
    /// Standardwert: ")"
    /// </summary>
    public string altParClose = ")";

    /// <summary>
    /// Dieser Bitvektor speichert jene Symbole, die einen Block erffnen
    /// ffnende Klammern.
    /// Standardwert: [Open, Assign]
    /// </summary>
    public BitSet openSet = new BitSet(Open,Assign);

    /// <summary>
    /// Dieser Bitvektor speichert jene Symbole, die einen Block beschlieen, z.B. 
    /// schlieende Klammern.
    /// Standardwert: [Close, Delimiter]
    /// </summary>
    public BitSet closeSet = new BitSet(Close,Delimiter);

    /// <summary>
    /// Dieser Bitvektor speichert Symbole, die ausgeblendet werden sollen.
    /// Mgliche Anwendungen: 
    /// - Verstecken von Kommentaren
    /// - Extraktion der EBNF-Grammatik aus der ATG
    /// Standardwert: []
    /// </summary>
    public BitSet hideSet = new BitSet();

    /// <summary>
    /// Enthlt jene Symbole, mit denen die rechte Spalte berlappen darf.
    /// Falls die rechte Spalte mit keinen Symbolen berlappen soll, wird der Bitvektor gelscht.
    /// Falls sie mit allen berlappen darf, wirden alle Bits gesetzt.
    /// Auch in diesem Fall ist gewhrleistet, dass innerhalb einer Zeile keine Symbole einander berschneiden.
    /// Standardwert: [OuterComment, Prolog, Epilog]
    /// </summary>
    public BitSet overlapSet = new BitSet(OuterComment,Prolog,Epilog);

    /// <summary>
    /// Dieser Array enthlt die Abstandswerte zwischen einzelnen Symbolen.
    /// 1. Dimension: Symbol
    /// 2. Dimension: Folgesymbol
    /// Wert: Abstand in Zeichen
    /// Standardwert: Siehe Klassendokumentation
    /// </summary>
    readonly sbyte[,] spaceMatrix =
    {
      { 0, 0, 2, 2, 2, 0, 0, 0, 0, 2, 0, 2, 2, 0, 0 }, // INDENT
      { 0, 1, 1, 1, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 0 }, // Production
      { 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0 }, // Symbol
      { 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0 }, // Literal
      { 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0 }, // Open
      { 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0 }, // Close
      { 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0 }, // Assign
      { 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0 }, // Separator
      { 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1 }, // Delimiter
      { 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0 }, // InnerComment
      { 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1 }, // OuterComment
      { 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0 }, // Attribute
      { 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 }, // Resolver
      { 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Prolog
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }  // Epilog
    };
    
    /// <summary>
    /// Dieser Array speichert das Verhalten beim Zeilenumbruch zwischen einzelnen Symbolen.
    /// 1. Dimension: Symbol
    /// 2. Dimension: Folgesymbol
    /// Wert: Nesting Level als Schranke fr einen Umbruch  oder BREAK_DEFAULT oder BREAK_NEVER
    /// Standardwert: Siehe Klassendokumentation
    /// </summary>
    readonly sbyte[,] breakMatrix =
    {
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // INDENT
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Production
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Symbol
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Literal 
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Open
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Close
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Assign
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Separator
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 }, // Delimiter
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // InnerComment
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 }, // OuterComment
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Attribute
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Resolver
      { 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // Prolog
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }  // Epilog
    };

    /// <summary>
    /// Diese Methode stellt den Abstand zwischen Symbolen ein.
    /// Die Werte fr Symbol und Folgesymbol mssen eine der defininierten Konstanten sein;
    /// ALL kann verwendet werden, um fr ein Symbol den Abstand zu allen Nachfolgesymbolen
    /// (oder Vorgngersymbolen) gleichzeitig einzustellen.
    /// 
    /// Beispiel:
    /// PutSpace(Symbol,Attribute,1); 
    /// Dieser Aufruf hat den Effekt, dass zwischen einem Symbol und dessen Attributsliste ein
    /// Abstand von einem Zeichen steht.
    /// 
    /// Beispiel:
    /// PutSpace(Open,ALL,2);
    /// Dieses Beispiel bewirkt, dass einer ffnenden Klammer immer ein Abstand von 
    /// zwei Zeichen folgt.   
    /// </summary>
    /// <param name="symbol">Symbol (alle auer Action und LineNumber) oder INDENT, falls der Einzug gesetzt werden soll. </param>
    /// <param name="follow">Folgesymbol (alle auer Action und LineNumber)</param>
    /// <param name="val">Abstand in Zeichen</param>
    public void PutSpace(int symbol, int follow, int val)
    {
      SetMatrix(spaceMatrix,symbol,follow,val);
    }

    /// <summary>
    /// Diese Methode steuert das Verhalten bei einem Zeilenumbruch zwischen zwei Symbolen.
    /// Die Werte fr Symbol und Folgesymbol mssen eine der defininierten Konstanten sein;
    /// ALL kann verwendet werden, um fr ein Symbol das Verhalten zu allen Nachfolgesymbolen
    /// (oder Vorgngersymbolen) gleichzeitig einzustellen.
    /// 
    /// Beispiel:
    /// PutBreak(Symbol,Attribute,BREAK_NEVER);
    /// Dies bewirkt, dass die Attributsliste eines Symbols niemals in die nchste Zeile gebrochen
    /// wird.
    /// 
    /// Beispiel:
    /// PutBreak(ALL,Delimiter,BREAK_ALWAYS);
    /// Diese Anwendung demonstriert, wie es mglich ist, zu erwingen, dass ein '.' immer in
    /// die nchste Zeile umgebrochen wird.
    /// </summary>
    /// <param name="symbol">Symbol (alle auer Action und LineNumber)</param>
    /// <param name="follow">Folgesymbol (alle auer Action und LineNumber)</param>
    /// <param name="val">BREAK_DEFAULT, BREAK_ALWAYS oder BREAK_NEVER</param>
    public void PutBreak(int symbol, int follow, int val)
    {
      SetMatrix(breakMatrix,symbol,follow,val);
    }

    /// <summary>
    /// Hilfesmethode fr PutSpace() und PutBreak().
    /// </summary>
    /// <param name="matrix"></param>
    /// <param name="symbol"></param>
    /// <param name="follow"></param>
    /// <param name="val"></param>
    void SetMatrix(sbyte[,] matrix, int symbol, int follow, int val)
    {
      int symmax = symbol;
      int folmax = follow;

      if (symbol == ALL)
      {
        symbol = MIN_SYMBOL;
        symmax = MAX_SYMBOL;
      }
      if (follow == ALL)
      {
        follow = MIN_SYMBOL;
        folmax = MAX_SYMBOL;
      }

      for (int i = symbol; i <= symmax; i++)
      {
        for (int j = follow; j <= folmax; j++)
        {
          matrix[i,j] = (sbyte)val;
        }
      }
    }

    /// <summary>
    /// Gibt den Abstand zwischwen zwei Symbolen zurck.
    /// </summary>
    /// <param name="symbol">Symbol</param>
    /// <param name="follow">Folgesymbol</param>
    /// <returns>Abstand</returns>
    public int GetSpace(int symbol, int follow)
    {
      return spaceMatrix[symbol,follow];
    }

    /// <summary>
    /// Gibt den Umbruch zwischen zwei Symbolen zurck.
    /// </summary>
    /// <param name="symbol">Symbol</param>
    /// <param name="follow">Folgesymbol</param>
    /// <returns>Umbruch</returns>
    public int GetBreak(int symbol, int follow)
    {
      return breakMatrix[symbol,follow];
    }
  }
}
