using System;
using System.Collections;
using System.IO;
using System.Text;

namespace Printy.Library
{
	/// <summary>
	/// Eine Erweiterung von TextBuilder, die Nonterminalsymbole mit Produktionen verlinkt.
	/// Ein HyperTextBuilder erstellt ein valides XHTML 1.0-Dokument (DOCTYPE XHTML 1.0 strict).
	/// Die Spaltenformatierungen werden von TextBuilder bernommen und in ein pre-Element
	/// eingebettet.
	/// Der HTB bietet auch Methoden zum Erstellen einer Overview-Frame (linke Frame mit allen Produktionen)
	/// und Index-Frameset (enthlt Overview- und Content-Frame).
	/// </summary>
	public class HyperTextBuilder : TextBuilder
	{
    /// <summary>
    /// Doctype: XHTML 1.0 Strict fr die Content-Frame
    /// </summary>
    const string DOCTYPE_STRICT =
      "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">";

    /// <summary>
    /// Doctype: XHTML 1.0 Transitional fr die Overview-Frame
    /// Transitional ist notwendig, weil das "base target"-Element bentigt wird, um in die rechte
    /// Frame zu linken
    /// </summary>
    const string DOCTYPE_TRANSITIONAL =
      "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">";

    /// <summary>
    /// Doctype: XHTML 1.0 Frameset fr die Index
    /// </summary>
    const string DOCTYPE_FRAMESET =
      "<!DOCTYPE html PUBLIC  \"-//W3C//DTD XHTML 1.0 Frameset//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd\">";

    /// <summary>
    /// Less Than
    /// </summary>
    const char LT = '<';

    /// <summary>
    /// Greater Than
    /// </summary>
    const char GT = '>';

    /// <summary>
    /// Ampersand
    /// </summary>
    const char AMP = '&';

    /// <summary>
    /// Diese Methode fhrt ein XML-Escape durch; d.h. dass alle Vorkommen von '<', '>' und '&'
    /// werden durch adquate XML-Entities ersetzt werden.
    /// </summary>
    /// <param name="text">Text</param>
    /// <returns>Text mit HTML-Entities</returns>
    static String Escape(String text)
    {
      StringBuilder result = new StringBuilder(text.Length);
      for (int i = 0; i < text.Length; i++)
      {
        char ch = text[i]; 
        switch (ch)
        {
          case LT:  result.Append("&lt;");  break;
          case GT:  result.Append("&gt;");  break;
          case AMP: result.Append("&amp;"); break;
          default:  result.Append(ch);      break;
        }
      }

      return result.ToString();
    }

    /// <summary>
    /// Erstellt einen Link auf eine Produktion.
    /// </summary>
    /// <param name="contentName">Name der Content-Frame</param>
    /// <param name="text">Text der Produktion</param>
    /// <returns>HTML-Linktext</returns>
    static string ProdLink(string contentName, string text)
    {
      return ProdLink(contentName,text,text);
    }

    /// <summary>
    /// Erstellt einen Link auf eine Produktion.
    /// </summary>
    /// <param name="contentName">Name der Content-Frame</param>
    /// <param name="text">Text der Produktion</param>
    /// <param name="result">bisheriges Ergebnis</param>
    /// <returns>HTML-Linktext</returns>
    static string ProdLink(string contentName, string text, string result)
    {
      return "<a class=\"ref_" + text + "\" href=\"" + contentName + "#" + text + "\">" + result + "</a>"; 
    }

    /// <summary>
    /// Erstellt den Link zu einer CSS-Datei.
    /// </summary>
    /// <param name="styleSheet">CSS-Dateiname</param>
    /// <returns>HTML-Linkelement</returns>
    static string StyleLink(string styleSheet)
    {
      return "<link rel=\"stylesheet\" href=\"" + styleSheet + "\"/>";
    }

    /// <summary>
    /// Schreibt den Anfang eines XHTML-Dokuments.
    /// </summary>
    /// <param name="writer">Ziel der Ausgaben</param>
    /// <param name="docType">DOCTYPE</param>
    /// <param name="grammarName">Name der Grammatik</param>
    static void WriteDocumentStartTo(TextWriter writer, string docType, string grammarName)
    {
      writer.WriteLine(docType);
      writer.Write("<html><head><title>Printy/");
      writer.Write(grammarName);
      writer.WriteLine("</title>");
      writer.WriteLine("<meta http-equiv=\"content-type\" content=\"text/html; charset=ISO-8859-1\"/>");
      writer.WriteLine("<meta name=\"generator\" content=\"Printy\"/>");
    }

    /// <summary>
    /// HTML-Konfiguration
    /// </summary>
    protected internal readonly HyperConfig htmlConfig;

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="grammar">Grammar-Knoten</param>
    public HyperTextBuilder(Grammar grammar) : this(grammar,new Config(),new HyperConfig())
    {}

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="grammar">Grammar-Knoten</param>
    /// <param name="config">Config-Objekt</param>
    public HyperTextBuilder(Grammar grammar, Config config) : this(grammar,config,new HyperConfig())
    {}

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="grammar">Grammar-Knoten</param>
    /// <param name="config">Config-Objekt</param>
    /// <param name="htmlConfig">HyperConfig-Objekt</param>
    public HyperTextBuilder(Grammar grammar, Config config, HyperConfig htmlConfig) : base(grammar,config)
    {
      this.htmlConfig = htmlConfig;
    }
      
    /// <summary>
    /// Diese Methode ist zentral fr die Erstellung der XHTML-Texte.
    /// </summary>
    /// <param name="text">Text</param>
    /// <param name="type">Typ</param>
    /// <returns>den Text, in XHTML-Elemente eingebettet</returns>
    protected override string Preprocess(string text, int type)
    {
      string className = null;
      string result = ((type == Config.Production) || type == Config.Symbol) ?  text : Escape(text);

      className = htmlConfig.GetStyleClass(text,type);
      if ((className == null) && htmlConfig.styleSet[type])
      {
        className = Config.Names[type];
      }
      if (className != null)
      {
        result = "<span class=\"" + className + "\">" + result + "</span>";
      }

      switch (type)
      {
        case Config.Production:
        {
          result = "<a id=\"" + text + "\">" + result + "</a>";
          break;
        }

        case Config.Symbol:
        {
          if (htmlConfig.createLinks && (grammar.nts[text] != null)) // Symbol ist ein NTS => verlinken
          {
            result = ProdLink("",text,result);
          }
          break;
        }
      }

      return result;
    }

    /// <summary>
    /// Ruft TextBuilder.WriteContentTo() zwischen HTML-Vor- und Nachspann auf.
    /// </summary>
    /// <param name="writer">Ziel der Ausgaben</param>
    public override void WriteContentTo(TextWriter writer)
    {
      WriteContentStartTo(writer);
      base.WriteContentTo(writer);
      WriteContentEndTo(writer);
    }

    /// <summary>
    /// Gibt den Start des HTML-Dokuments aus.
    /// </summary>
    /// <param name="writer">Ziel der Ausgaben</param>
    public void WriteContentStartTo(TextWriter writer)
    {
      WriteDocumentStartTo(writer,DOCTYPE_STRICT,grammar.text);
      if (htmlConfig.styleSheet != null)
      {
        writer.WriteLine(StyleLink(htmlConfig.styleSheet));
      }
      writer.WriteLine("</head><body><pre>");
    }

    /// <summary>
    /// Gibt das Ende des HTML-Dokuments aus.
    /// </summary>
    /// <param name="writer">Ziel der Ausgaben</param>
    public void WriteContentEndTo(TextWriter writer)
    {
      writer.WriteLine("</pre></body></html>");
    }

    /// <summary>
    /// Schreibt eine HTML-Frame, die eine Liste alle Produktion mit Links in die Content-Frame enthlt.
    /// Diese Methode darf auch vor Printer.Print() aufgerufen werden.
    /// </summary>
    /// <param name="writer">Ziel der Ausgaben</param>
    /// <param name="contentName">Name der Content-Frame</param>
    /// <param name="sort">Sollen Produktionen sortiert ausgegeben werden?</param>
    public virtual void WriteOverviewTo(TextWriter writer, string contentName, bool sort)
    {
      WriteDocumentStartTo(writer,DOCTYPE_TRANSITIONAL,grammar.text);
      writer.WriteLine("<base target=\"content\"/>");
      if (htmlConfig.styleSheet != null)
      {
        writer.WriteLine(StyleLink(htmlConfig.styleSheet));
      }
      writer.WriteLine("</head><body><div>");

      if (sort) // sortiert ausgeben
      {
        for (IDictionaryEnumerator e = grammar.nts.GetEnumerator(); e.MoveNext();)
        {
          writer.Write(ProdLink(contentName,(string)e.Key));
          writer.WriteLine("<br/>");
        }
      }
      else // in originaler Reihenfolge ausgeben
      {
        for (Node p = grammar.sub; p != null; p = p.next)
        {
          writer.Write(ProdLink(contentName,p.text));
          writer.WriteLine("<br/>");
        }
      }
      writer.WriteLine("</div></body></html>");
    }

    /// <summary>
    /// Schreibt ein HTML-Frameset, das die Overview-Frame und Content-Frame enthlt.
    /// Diese Methode darf auch vor Printer.Print() aufgerufen werden.
    /// </summary>
    /// <param name="writer">Ziel der Ausgaben</param>
    /// <param name="overviewName">Name der Overview-Frame</param>
    /// <param name="contentName">Name der Content-Frame</param>
    /// <param name="left">soll die Overview-Frame links stehen?</param>
    public virtual void WriteIndexTo(TextWriter writer, string overviewName, string contentName, bool left)
    {
      WriteDocumentStartTo(writer,DOCTYPE_FRAMESET,grammar.text);
      if (left)
      {
        writer.WriteLine("</head><frameset cols=\"20%,*\">");
        writer.WriteLine("<frame name=\"overview\" src=\"" + overviewName + "\"/>");
        writer.WriteLine("<frame name=\"content\" src=\"" + contentName + "\"/>");
      }
      else
      {
        writer.WriteLine("</head><frameset cols=\"*,20%\">");
        writer.WriteLine("<frame name=\"content\" src=\"" + contentName + "\"/>");
        writer.WriteLine("<frame name=\"overview\" src=\"" + overviewName + "\"/>");
      }
      writer.WriteLine("</frameset></html>");
    }
	}
}
