using System;
using System.IO;

namespace Printy.Library
{
	/// <summary>
	/// Formatiert einen ATG-Syntaxbaum und schreibt das Ergebnis in einen TextBuilder.
	/// Alle Methoden sind rekursiv implementiert.
	/// Diese Klasse legt den Ablauf fest; sie bentzt einen TextBuilder (oder eine Subklasse)
	/// fr Formatierungen.
	/// </summary>
	public class Printer
	{
    /// <summary>
    /// Carriage return \13
    /// </summary>
    const int CR = '\r';

    /// <summary>
    /// Line feed \10
    /// </summary>
    const int LF = '\n';

    /// <summary>
    /// Leerzeichen
    /// </summary>
    const int SPACE = ' ';

    /// <summary>
    /// Ende der Eingabe (fr Tokenize)
    /// </summary>
    const int END = -1;

    /// <summary>
    /// Gibt das Zeichen oder END zurck
    /// </summary>
    /// <param name="text">Text</param>
    /// <param name="index">Index des Zeichens</param>
    /// <returns>text[index] falls index gltig ist, END sonst</returns>
    static int GetCh(string text, int index)
    {
      return (index < text.Length) ? text[index] : END;
    }

    /// <summary>
    /// Zerlegt die Zeichenkette in Zeilen, die in einem Array abgelegt werden.
    /// Es werden keine CRLF Steuerzeichen ausgegeben; es wird CR, LF und CRLF als Zeilenseparator
    /// erkannt.
    /// Alle Whitespace-Zeichen vor dem Zeilenwechsel werden abgeschnitten.
    /// </summary>
    /// <param name="text">Text mit Steuerzeichen</param>
    /// <param name="result">Enthlt nach dem Aufruf die einzelnen Zeilen.
    /// Der Array muss gro genug sein, um alle Zeilen aufnehmen zu knnen.
    /// </param>
    static string[] Tokenize(string text, string[] result)
    {
      int index = 0;          // Zeilenindex
      int pos = 0;            // Position im Text
      int start = 0;          // Start der aktuellen Zeile
      int end = 0;            // Ende der aktuellen Zeile
      int ch;                 // aktuelles Zeichen

      while ((ch = GetCh(text,pos)) != END)
      {
        while ((ch == CR) || (ch == LF))
        {
          result[index++] = text.Substring(start,end - start);
          if (ch == CR) ch = GetCh(text,++pos);
          if (ch == LF) ch = GetCh(text,++pos);
          start = end = pos;
        }
        pos++;
        if (ch > SPACE) end = pos;
      }
      result[index] = (start < end) ? text.Substring(start,end - start) : null;
      return result;
    }

    /// <summary>
    /// Konfigurationsobjekt
    /// </summary>
    Config config;

    /// <summary>
    /// Zeilenpuffer
    /// </summary>
    TextBuilder builder;

    /// <summary>
    /// Zeilenpuffer
    /// </summary>
    string[] buffer;

    /// <summary>
    /// aktuelle Zeilennummer des Eingabedokuments
    /// </summary>
    int line;

    /// <summary>
    /// Gibt soviele Zeilenwechsel aus, bis newLine erreicht ist.
    /// </summary>
    /// <param name="newLine"></param>
    void PrintBreak(int newLine)
    {
      if (!config.noBreaks && (line > 0))
      {
        int n = newLine - line;
        if (n > 0)
        {
          builder.CloseLine();
        }
        for (int i = n; i > 0; i--)
        {
          builder.NewLine();
        }
      }
      line = newLine;
    }

    /// <summary>
    /// Gibt den angegebenen Text aus.
    /// </summary>
    /// <param name="text">Text</param>
    /// <param name="type">Typ des Textes</param>
    /// <param name="level">Ebene</param>
    void PrintText(string text, int type, int level)
    {
      if (!config.hideSet[type])
      {
        AppendText(text,type,level);
      }
    }

    /// <summary>
    /// Gibt mehrere Zeilen aus.
    /// Dies kann ein mehrzeiliger Kommentar, ein Resolver oder eine Attributsliste sein.
    /// </summary>
    /// <param name="text">Text (mit Steuerzeichen)</param>
    /// <param name="lines">Anzahl der Zeilen - 1</param>
    /// <param name="type">Typ des Textes</param>
    /// <param name="level">Ebene</param>
    void PrintTextLines(string text, int lines, int type, int level)
    {
      if (!config.hideSet[type])
      {
        string[] buf = Tokenize(text,buffer);
        for (int i = 0; i <= lines; i++)
        {
          if (i > 0)
          {
            builder.CloseLine();
            builder.NewLine();
          }
          AppendText(buf[i],type,level);
        }
      }
      line += lines;
    }

    /// <summary>
    /// Gibt eine semantische Aktion aus.
    /// </summary>
    /// <param name="text">Text der semantischen Aktion</param>
    void PrintAction(string text)
    {
      if (!config.hideSet[Config.Action])
      {
        builder.PutAction(text);
        builder.CloseLine();
      }
    }

    /// <summary>
    /// Gibt eine mehrzeilige semantische Aktion aus.
    /// </summary>
    /// <param name="text">Mehrzeiliger Text (mit Steuerzeichen)</param>
    /// <param name="lines">Anzahl der Zeilen - 1</param>
    /// <param name="col">Spalte der ffnenden Klammer "(."</param>
    void PrintActionLines(string text, int lines, int col)
    {
      if (!config.hideSet[Config.Action])
      {
        string[] buf = Tokenize(text,buffer);
        for (int i = 0; i <= lines; i++)
        {
          string action = buf[i];
          if (action != null)
          {
            int length = Math.Min(action.Length,col - 1);
            int start; // Start der Zeile finden
            for (start = 0; (start < length) && (action[start] <= SPACE); start++); 
            builder.PutAction(action.Substring(start));
          }
          if (i == 0)
          {
            builder.CloseLine();
          }
          if (i < lines)
          {
            builder.NewLine();
          }
        }
      }
      line += lines;
    }

    /// <summary>
    /// Beschliet eine semantische Aktion mit einem Zeilenumbruch.
    /// </summary>
    void CloseAction()
    {
      if (!config.hideSet[Config.Action])
      {
        builder.NewLine();
        line++;
      }
    }

    /// <summary>
    /// Gibt ein Token sowie alle daran anschlieenden Kommentare aus.
    /// </summary>
    /// <param name="token">Token</param>
    /// <param name="type">Typ des Tokens</param>
    /// <param name="level">Level</param>
    void PrintToken(Token token, int type, int level)
    {
      int commentType = (type == Config.Delimiter) ? Config.OuterComment : Config.InnerComment;

      for (Token t = token; t != null; t = t.comment)
      {
        PrintBreak(t.line);
        if (t.lines > 0) // mehrzeiliges Token ausgeben (Kommentar)
        {
          PrintTextLines(t.val,t.lines,type,level);
        }
        else
        {
          PrintText(t.val,type,level);
        }
        type = commentType;
      }
    }

    /// <summary>
    /// Gibt den angegebenen Text aus.
    /// </summary>
    /// <param name="text">Der auszugebende Text; falls null, wird nichts ausgegeben</param>
    /// <param name="type">Typ des Textes</param>
    /// <param name="level">Ebene</param>
    void AppendText(string text, int type, int level)
    {
      if (text != null)
      {
        if (!builder.CanAppendText(text,type,level)) // Neue Zeile
        {
          builder.CloseLine();
          builder.NewLine();
        }
        builder.AppendText(text,type);
      }
    }

    /// <summary>
    /// Gibt einen Knoten aus.
    /// </summary>
    /// <param name="node">Knoten</param>
    /// <param name="level">Ebene</param>
    void PrintOneNode(Node node, int level)
    {
      PrintToken(node.sep,Config.Separator,level);
      switch (node.type)
      {
        case Node.Grammar:
        {
          PrintNode(node.sub,level + 1);
          break;
        }

        case Node.Production:
        {
          PrintToken(node.sym,Config.Production,level);
          PrintNode(node.meta,level);
          PrintToken(node.open,Config.Assign,level);
          PrintNode(node.sub,level + 1);
          PrintToken(node.close,Config.Delimiter,level);
          break;
        }

        case Node.Action:
        {
          PrintBreak(node.open.line);
          if (node.close.line > node.open.line)
          {
            PrintActionLines(node.text,node.close.line - node.open.line,node.open.col);
          }
          else
          {
            PrintAction(node.text);
          }
          CloseAction();
          PrintToken(node.close.comment,Config.InnerComment,level);
          break;
        }

        case Node.Resolver:
        case Node.Attribute:
        {
          int type = (node.type == Node.Resolver) ? Config.Resolver : Config.Attribute;
          PrintBreak(node.open.line);
          if (node.close.line > node.open.line)
          {
            PrintTextLines(node.text,node.close.line - node.open.line,type,level);
          }
          else
          {
            PrintText(node.text,type,level);
          }
          PrintToken(node.close.comment,Config.InnerComment,level);
          break;
        }

        default: // Group, Symbol, Literal, Epsilon
        {
          int type = (node.type == Node.Symbol) ? Config.Symbol : Config.Literal;
          PrintToken(node.open,Config.Open,level);
          PrintNode(node.sub,level + 1);
          PrintToken(node.sym,type,level);
          PrintNode(node.meta,level);
          PrintToken(node.close,Config.Close,level);
          break;
        }
      }
    }

    /// <summary>
    /// Gibt einen Knoten und alle darauf folgenden Knoten aus.
    /// </summary>
    /// <param name="node">Knoten</param>
    /// <param name="level">Ebene</param>
    void PrintNode(Node node, int level)
    {
      bool altPar = (node != null) && (node.alt != null) && Config.TestLevel(level,config.altPar);

      if (altPar) // ffnende Klammer fr Alternativenlisten
      {
        if (node.sym != null) 
        {
          PrintBreak(node.sym.line);
        }
        else if (node.open != null) 
        {
          PrintBreak(node.open.line);
        }
        PrintText(config.altParOpen,Config.Open,level++);
      }

      for (Node p = node; p != null; p = p.alt)
      {
        for (Node q = p; q != null; q = q.next)
        {
          PrintOneNode(q,level);
        }
      }
      
      if (altPar) // Schlieende Klammer fr Alternativenlisten
      {
        PrintText(config.altParClose,Config.Close,--level);
      }
    }

    /// <summary>
    /// Schreibt den Pro- oder Epilog.
    /// </summary>
    /// <param name="node">Log-Knoten</param>
    /// <param name="type">Config.Prolog oder Config.Epilog</param>
    void PrintLog(Node node, int type)
    {
      int openLine = (node.open != null) ? node.open.line : 1;
      int closeLine = node.close.line;
      PrintBreak(openLine);
      if (closeLine > openLine) 
      {
        PrintTextLines(node.text,closeLine - openLine,type,1);
      }
      else
      {
        PrintText(node.text,type,1);
      }
    }

    /// <summary>
    /// Gibt die gesamte Grammatik in den TextBuilder aus.
    /// Diese Methode traversiert den Syntaxgraphen und legt den formatierten Text
    /// im TextBuilder ab. Um den Text auszugeben, kann TextBuilder.WriteContentTo()
    /// verwendet werden.
    /// </summary>
    /// <param name="builder">TextBuilder-Objekt (darf nicht null sein)</param>
    public void Print(TextBuilder builder)
    {
      Print(builder,builder.grammar);
    }

    /// <summary>
    /// Gibt einen bestimmten Knoten einer Grammatik aus.
    /// Diese Methode traversiert den Syntaxgraphen beginnend beim angegebenen
    /// Knoten und legt den formatierten Text im TextBuilder ab. 
    /// Um den Text auszugeben, kann TextBuilder.WriteContentTo()
    /// verwendet werden.
    /// </summary>
    /// <param name="builder">TextBuilder-Objekt (darf nicht null sein)</param>
    /// <param name="node">Node-Objekt (darf nicht null sein)</param>
    public void Print(TextBuilder builder, Node node)
    {
      this.builder = builder;
      config = builder.config;
      buffer = new string[builder.grammar.maxLines + 1];
      line = 0;

      if (node == builder.grammar) // gesamte Grammatik inklusive Prolog und Epilog
      {
        PrintLog(node.meta,Config.Prolog);
        PrintOneNode(node,node.level);
        PrintLog(node.meta.next,Config.Epilog);
      }
      else // nur den Knoten ausgeben
      {
        PrintOneNode(node,node.level);
      }
      builder.CloseLine();
    }
	}
}
